/* -*- mode: C; c-file-style: "gnu"; indent-tabs-mode: nil; -*- */

/**
 * SECTION:shell-mag-brightness-effect
 * @short_description: Brightness effect for the magnifier.
 * @see_also: #ClutterEffect, #ClutterOffscreenEffect
 *
 * #ShellMagBrightnessEffect is a sub-class of #ClutterEffect that changes the
 * overall brightness of a clutter actor.
 */

#include "shell-mag-brightness-effect.h"

#include <string.h>
#include <cogl/cogl.h>

struct _ShellMagBrightnessEffect
{
  ClutterOffscreenEffect parent_instance;

  /* a back pointer to our actor, so that we can query it */
  ClutterActor *actor;

  CoglHandle shader;
  CoglHandle program;

  /* Brightness changes. */
  ClutterColor brightness;

  gint tex_uniform;
  gint brightness_uniform;

  guint is_attached : 1;
};

struct _ShellMagBrightnessEffectClass
{
  ClutterOffscreenEffectClass parent_class;
};

/* Brightness effects in GLSL.
 */
static const gchar *magnifier_brightness_glsl_shader =
"uniform sampler2D tex;\n"
"uniform vec3 brightness;\n"
"\n"
"void main ()\n"
"{\n"
"  vec4 color = cogl_color_in * texture2D (tex, vec2 (cogl_tex_coord_in[0].xy));\n"
"  vec3 effect = vec3 (color);\n"
"\n"
"  effect = clamp (effect + brightness, 0.0, 1.0);\n"
"\n"
"\n"
"  cogl_color_out = vec4 (effect, color.a);\n"
"}\n";

/* No brightness change. */
static const ClutterColor same_brightness = { 0x7f, 0x7f, 0x7f, 0xff };

enum
{
  PROP_0,

  PROP_BRIGHTNESS,

  PROP_LAST
};

static GParamSpec *obj_props[PROP_LAST];

G_DEFINE_TYPE (ShellMagBrightnessEffect,
               shell_mag_brightness_effect,
               CLUTTER_TYPE_OFFSCREEN_EFFECT);

static gboolean
shell_mag_brightness_effect_pre_paint (ClutterEffect *effect)
{
  ShellMagBrightnessEffect *self = SHELL_MAG_BRIGHTNESS_EFFECT (effect);
  ClutterEffectClass *parent_class;

  if (self->shader == COGL_INVALID_HANDLE)
    return FALSE;

  if (!clutter_actor_meta_get_enabled (CLUTTER_ACTOR_META (effect)))
    return FALSE;

  if (self->actor == NULL)
    return FALSE;

  if (self->program == COGL_INVALID_HANDLE)
    self->program = cogl_create_program ();

  if (!self->is_attached)
    {
      g_assert (self->shader != COGL_INVALID_HANDLE);
      g_assert (self->program != COGL_INVALID_HANDLE);

      cogl_program_attach_shader (self->program, self->shader);
      cogl_program_link (self->program);

      cogl_handle_unref (self->shader);

      self->is_attached = TRUE;

      self->tex_uniform =
        cogl_program_get_uniform_location (self->program, "tex");
      self->brightness_uniform =
        cogl_program_get_uniform_location (self->program, "brightness");
    }

  parent_class = CLUTTER_EFFECT_CLASS (shell_mag_brightness_effect_parent_class);
  return parent_class->pre_paint (effect);
}

static void
shell_mag_brightness_effect_paint_target (ClutterOffscreenEffect *effect)
{
  ShellMagBrightnessEffect *self = SHELL_MAG_BRIGHTNESS_EFFECT (effect);
  ClutterOffscreenEffectClass *parent;
  CoglHandle material;

  if (self->program == COGL_INVALID_HANDLE)
    goto out;

  if (self->tex_uniform > -1)
    cogl_program_set_uniform_1i (self->program, self->tex_uniform, 0);

  if (self->brightness_uniform > -1)
    {
      float brightness[3] = {
        (self->brightness.red / 127.0) - 1.0,
        (self->brightness.green / 127.0) - 1.0,
        (self->brightness.blue / 127.0) - 1.0
      };
      cogl_program_set_uniform_float (self->program, self->brightness_uniform,
                                      3, 1,
                                      brightness);
    }

  material = clutter_offscreen_effect_get_target (effect);
  cogl_material_set_user_program (material, self->program);

out:
  parent = CLUTTER_OFFSCREEN_EFFECT_CLASS (shell_mag_brightness_effect_parent_class);
  parent->paint_target (effect);
}

static void
shell_mag_brightness_effect_set_actor (ClutterActorMeta *meta,
                                       ClutterActor *actor)
{
  ShellMagBrightnessEffect *self = SHELL_MAG_BRIGHTNESS_EFFECT (meta);
  ClutterActorMetaClass *parent;

  g_return_if_fail (actor == NULL || CLUTTER_IS_ACTOR (actor));

  if (self->shader == COGL_INVALID_HANDLE)
    {
      clutter_actor_meta_set_enabled (meta, FALSE);
      return;
    }

  parent = CLUTTER_ACTOR_META_CLASS (shell_mag_brightness_effect_parent_class);
  parent->set_actor (meta, actor);

  /* keep a back pointer here, to avoid going through the ActorMeta */
  self->actor = clutter_actor_meta_get_actor (meta);
}

static void
shell_mag_brightness_effect_dispose (GObject *gobject)
{
  ShellMagBrightnessEffect *self = SHELL_MAG_BRIGHTNESS_EFFECT (gobject);

  if (self->program != COGL_INVALID_HANDLE)
    {
      cogl_handle_unref (self->program);

      self->program = COGL_INVALID_HANDLE;
      self->shader = COGL_INVALID_HANDLE;
    }

  self->actor = NULL;

  G_OBJECT_CLASS (shell_mag_brightness_effect_parent_class)->dispose (gobject);
}

static void
shell_mag_brightness_effect_set_property (GObject      *gobject,
                                          guint        prop_id,
                                          const GValue *value,
                                          GParamSpec   *pspec)
{
  ShellMagBrightnessEffect *effect = SHELL_MAG_BRIGHTNESS_EFFECT (gobject);

  switch (prop_id)
    {
    case PROP_BRIGHTNESS:
        shell_mag_brightness_effect_set_brightness (effect,
                                                    clutter_value_get_color (value));
        break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, prop_id, pspec);
      break;
    }
}

static void
shell_mag_brightness_effect_get_property (GObject    *gobject,
                                         guint       prop_id,
                                         GValue     *value,
                                         GParamSpec *pspec)
{
  ShellMagBrightnessEffect *effect = SHELL_MAG_BRIGHTNESS_EFFECT (gobject);
  ClutterColor *aColor = clutter_color_new (127, 127, 127, 255);
  switch (prop_id)
    {
    case PROP_BRIGHTNESS:
        shell_mag_brightness_effect_get_brightness (effect, aColor);
        clutter_value_set_color (value, aColor);
        break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, prop_id, pspec);
      break;
    }
    clutter_color_free (aColor);
}

static void
shell_mag_brightness_effect_class_init (ShellMagBrightnessEffectClass *klass)
{
  ClutterEffectClass *effect_class = CLUTTER_EFFECT_CLASS (klass);
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  ClutterOffscreenEffectClass *offscreen_class;
  ClutterActorMetaClass *meta_class = CLUTTER_ACTOR_META_CLASS (klass);

  offscreen_class = CLUTTER_OFFSCREEN_EFFECT_CLASS (klass);
  offscreen_class->paint_target = shell_mag_brightness_effect_paint_target;

  meta_class->set_actor = shell_mag_brightness_effect_set_actor;

  effect_class->pre_paint = shell_mag_brightness_effect_pre_paint;

  gobject_class->set_property = shell_mag_brightness_effect_set_property;
  gobject_class->get_property = shell_mag_brightness_effect_get_property;
  gobject_class->dispose = shell_mag_brightness_effect_dispose;

  /**
   * ShellMagBrightnessEffect:brightness:
   *
   * The brightness change to apply to the actor
   *
   */
  obj_props[PROP_BRIGHTNESS] =
    clutter_param_spec_color ("brightness",
                              "Brightness",
                              "The brightness change to apply",
                              &same_brightness,
                              G_PARAM_READWRITE);

  g_object_class_install_properties (gobject_class, PROP_LAST, obj_props);

}

static void
shell_mag_brightness_effect_init (ShellMagBrightnessEffect *self)
{
  static CoglHandle shader = COGL_INVALID_HANDLE;

  if (shader == COGL_INVALID_HANDLE)
    {
      if (clutter_feature_available (CLUTTER_FEATURE_SHADERS_GLSL))
        {
          shader = cogl_create_shader (COGL_SHADER_TYPE_FRAGMENT);
          cogl_shader_source (shader, magnifier_brightness_glsl_shader);
          cogl_shader_compile (shader);
          if (!cogl_shader_is_compiled (shader))
            {
              gchar *log_buf = cogl_shader_get_info_log (shader);
              g_warning (G_STRLOC ":  Unable to compile the magnifier brightness effects shader: %s",
                         log_buf);
              g_free (log_buf);

              cogl_handle_unref (shader);
              shader = COGL_INVALID_HANDLE;
            }
        }
    }

  self->shader = shader;
  self->is_attached = FALSE;
  self->tex_uniform = -1;
  self->brightness_uniform = -1;
  self->brightness = same_brightness;

  if (shader != COGL_INVALID_HANDLE)
    cogl_handle_ref (self->shader);
}

/**
 * shell_mag_brightness_effect_new:
 *
 * Creates a new #ShellMagBrightnessEffect to be used with
 * clutter_actor_add_effect()
 *
 * Return value: the newly created #ShellMagBrightnessEffect or %NULL
 */
ClutterEffect *
shell_mag_brightness_effect_new (void)
{
  return g_object_new (SHELL_TYPE_MAG_BRIGHTNESS_EFFECT,
                       NULL);
}

/**
 * shell_mag_brightness_effect_set_brightness:
 * @effect: a #ShellMagBrightnessEffect
 * @brightness: ClutterColor governing the brightness change.
 *
 * Add each of the red, green, blue components of the @brightness to
 * the red, greeen, or blue component of the actor's colors.
 */
void
shell_mag_brightness_effect_set_brightness (ShellMagBrightnessEffect *effect,
                                            const ClutterColor       *brightness)
{
  g_return_if_fail (SHELL_IS_MAG_BRIGHTNESS_EFFECT (effect));
  if (clutter_color_equal (&effect->brightness, &brightness))
    return;

  effect->brightness = *brightness;

  if (effect->actor != NULL)
    clutter_actor_queue_redraw (effect->actor);
}

/**
 * shell_mag_brightness_effect_get_brightness:
 * @effect: a #ShellMagBrightnessEffect
 *
 * Retrieves the brightness value used by @effect
 */
void
shell_mag_brightness_effect_get_brightness (ShellMagBrightnessEffect *effect,
                                            ClutterColor             *brightness)
{
  g_return_if_fail (SHELL_IS_MAG_BRIGHTNESS_EFFECT (effect));
  g_return_if_fail (brightness != NULL);

  *brightness = effect->brightness;
}
